"""Graph database for storing harvested data."""
from neo4j import GraphDatabase
from neo4j.exceptions import ServiceUnavailable, ConstraintError
from typing import Tuple, Dict, Set
import logging
import json
from pathlib import Path
from os import environ
import boto3
import base64
from botocore.exceptions import ClientError
import ast

logger = logging.getLogger('metakb.database')
logger.setLevel(logging.DEBUG)


def _create_keys_string(entity, keys) -> str:
    """Create formatted string for requested keys if non-null in entity.
    :param Dict entity: entity to check against, eg a Disease or Statement
    :param Tuple keys: key names to check
    :return: formatted String for use in Cypher query
    """
    nonnull_keys = [f"{key}:${key}"
                    for key in keys if entity.get(key)]
    keys_string = ', '.join(nonnull_keys)
    return keys_string


class Graph:
    """Manage requests to graph datastore."""

    def __init__(self, uri: str = '', credentials: Tuple[str, str] = ('', '')):
        """Initialize Graph driver instance.
        :param str uri: address of Neo4j DB
        :param Tuple[str, str] credentials: tuple containing username and
            password
        """
        if 'METAKB_NORM_EB_PROD' in environ:
            secret = ast.literal_eval(self.get_secret())
            uri = f"bolt://{secret['host']}:{secret['port']}"
            credentials = (secret['username'], secret['password'])
        elif 'METAKB_DB_URL' in environ and 'METAKB_DB_USERNAME' in environ and 'METAKB_DB_PASSWORD' in environ:  # noqa: E501
            uri = environ['METAKB_DB_URL']
            credentials = (environ['METAKB_DB_USERNAME'],
                           environ['METAKB_DB_PASSWORD'])
        elif not (uri and credentials[0] and credentials[1]):
            # Local
            uri = "bolt://localhost:7687"
            credentials = ("neo4j", "admin")
        self.driver = GraphDatabase.driver(uri, auth=credentials)
        with self.driver.session() as session:
            session.write_transaction(self._create_constraints)

    def close(self):
        """Close Neo4j driver."""
        self.driver.close()

    def clear(self):
        """Debugging helper - wipe out DB."""
        def delete_all(tx):
            tx.run("MATCH (n) DETACH DELETE n;")
        with self.driver.session() as session:
            session.write_transaction(delete_all)

    def load_from_json(self, src_transformed_cdm: Path):
        """Load evidence into DB from given JSON file.
        :param Path src_transformed_cdm: path to file for a source's
            transformed data to common data model containing statements,
            propositions, variation/therapy/disease/gene descriptors,
            methods, and documents
        """
        logger.info(f"Loading data from {src_transformed_cdm}")
        with open(src_transformed_cdm, 'r') as f:
            items = json.load(f)
            self.add_transformed_data(items)

    @staticmethod
    def _create_constraints(tx):
        """Create unique property constraints for ID values"""
        for label in [
            'Gene', 'Disease', 'Therapy', 'Variation', 'GeneDescriptor',
            'TherapyDescriptor', 'DiseaseDescriptor',
            'VariationDescriptor', 'VariationGroup', 'Proposition',
            'Document', 'Statement', 'Method'
        ]:
            query = (
                f"CREATE CONSTRAINT {label.lower()}_id_constraint "
                f"IF NOT EXISTS ON (n:{label}) "
                "ASSERT n.id IS UNIQUE;"
            )
            try:
                tx.run(query)
            except ServiceUnavailable as exception:
                logging.error(f"Failed to generate ID property "
                              f"constraint for {label}.")
                raise exception

    def add_transformed_data(self, data: Dict):
        """Add set of data formatted per Common Data Model to DB.
        :param Dict data: contains key/value pairs for data objects to add
            to DB, including statements, propositions,
            variation/therapy/disease/gene descriptors, methods, and documents
        """
        added_ids = set()  # Used to keep track of IDs that are in statements
        with self.driver.session() as session:
            loaded_count = 0
            for ev in data.get('statements', []):
                self._get_ids_from_statement(ev, added_ids)
            for var_descr in data.get('variation_descriptors', []):
                if var_descr['id'] in added_ids:
                    gc = var_descr['gene_context']
                    if gc:
                        added_ids.add(gc)
            for method in data.get('methods', []):
                try:
                    session.write_transaction(self._add_method, method,
                                              added_ids)
                except ConstraintError:
                    logger.warning(f"{method['id']} exists already.")
                    continue
            for descriptor in ['therapy_descriptors', 'disease_descriptors',
                               'gene_descriptors']:
                for d in data.get(descriptor, []):
                    try:
                        session.write_transaction(
                            self._add_descriptor, d, added_ids
                        )
                    except ConstraintError:
                        logger.warning(f"{d['id']} exists already.")
                        continue
            for var_descr in data.get('variation_descriptors', []):
                try:
                    session.write_transaction(self._add_variation_descriptor,
                                              var_descr, added_ids)
                except ConstraintError:
                    logger.warning(f"{var_descr['id']} exists already.")
                    continue
            for doc in data.get('documents'):
                try:
                    session.write_transaction(self._add_document, doc)
                except ConstraintError:
                    logger.warning(f"{doc['id']} exists already.")
                    continue
            for proposition in data.get('propositions', []):
                try:
                    session.write_transaction(self._add_proposition,
                                              proposition)
                except ConstraintError:
                    logger.warning(f"{proposition['id']} exists already.")
                    continue
            for s in data.get('statements', []):
                loaded_count += 1
                try:
                    session.write_transaction(self._add_statement, s,
                                              added_ids)
                except ConstraintError:
                    logger.warning(f"{s['id']} exists already.")
            logger.info(f"Successfully loaded {loaded_count} statements.")

    @staticmethod
    def _add_method(tx, method: Dict, added_ids: Set[str]):
        """Add Method object to DB.
        :param Dict method: must include `id`, `label`, `url`,
            `version`, and `authors` values.
        :param set added_ids: IDs found in statements
        """
        method['version'] = json.dumps(method['version'])
        query = """
        MERGE (n:Method {id:$id, label:$label, url:$url,
            version:$version, authors: $authors});
        """
        if method['id'] in added_ids:
            try:
                tx.run(query, **method)
            except ServiceUnavailable as exception:
                logging.error(f"Failed to add Method object\nQuery: "
                              f"{query}\nAssertionMethod: {method}")
                raise exception

    @staticmethod
    def _add_descriptor(tx, descriptor: Dict, added_ids: Set[str]):
        """Add gene, therapy, or disease descriptor object to DB.
        :param Dict descriptor: must contain a `value` field with `type`
            and `<type>_id` fields. `type` field must be one of
            {'TherapyDescriptor', 'DiseaseDescriptor', 'GeneDescriptor'}
        :param set added_ids: IDs found in statements
        """
        if descriptor['id'] not in added_ids:
            return
        descr_type = descriptor['type']
        if descr_type == 'TherapyDescriptor':
            value_type = 'Therapy'
        elif descr_type == 'DiseaseDescriptor':
            value_type = 'Disease'
        elif descr_type == 'GeneDescriptor':
            value_type = 'Gene'
        else:
            raise TypeError(f"Invalid Descriptor type: {descr_type}")

        value_id = f"{value_type.lower()}_id"
        descr_keys = _create_keys_string(descriptor, ('id', 'label',
                                                      'description', 'xrefs',
                                                      'alternate_labels'))

        query = f'''
        MERGE (descr:{descr_type} {{ {descr_keys} }})
        MERGE (value:{value_type} {{ id:${value_id} }})
        MERGE (descr) -[:DESCRIBES]-> (value)
        '''
        try:
            tx.run(query, **descriptor)
        except ServiceUnavailable as exception:
            logging.error(f"Failed to add Descriptor object\nQuery: {query}\n"
                          f"Descriptor: {descriptor}")
            raise exception

    @staticmethod
    def _add_variation_descriptor(tx, descriptor_in: Dict,
                                  added_ids: Set[str]):
        """Add variant descriptor object to DB.
        :param Dict descriptor_in: must include a `value_id` field and a
            `value` object containing `type`, `state`, and `location` objects.
        :param set added_ids: IDs found in statements
        """
        descriptor = descriptor_in.copy()

        if descriptor['id'] not in added_ids:
            return

        # prepare value properties
        variation_type = descriptor['variation']['type']
        descriptor['variation'] = json.dumps(descriptor['variation'])
        # prepare descriptor properties
        expressions = descriptor.get('expressions')
        if expressions:
            for expression in expressions:
                syntax = expression['syntax'].split(':')[1]
                key = f"expressions_{syntax}"
                if key in descriptor:
                    descriptor[key].append(expression['value'])
                else:
                    descriptor[key] = [expression['value']]

        nonnull_keys = [_create_keys_string(descriptor,
                                            ('id', 'label', 'description',
                                             'xrefs', 'alternate_labels',
                                             'structural_type',
                                             'molecule_context',
                                             'expressions_transcript',
                                             'expressions_genomic',
                                             'expressions_protein',
                                             'vrs_ref_allele_seq'))]

        # handle extensions
        variant_groups = None
        extensions = descriptor.get('extensions')
        if extensions:
            for ext in extensions:
                name = ext['name']
                if name == 'variant_group':
                    variant_groups = ext['value']
                else:
                    descriptor[name] = json.dumps(ext['value'])
                    nonnull_keys.append(f"{name}:${name}")

        descriptor_keys = ', '.join(nonnull_keys)

        query = f"""
        MERGE (descr:VariationDescriptor
            {{ {descriptor_keys} }})
        MERGE (value:{variation_type}:Variation
            {{id:$variation_id,
              variation:$variation
              }})
        MERGE (gene_context:GeneDescriptor {{id:$gene_context}} )
        MERGE (descr) -[:DESCRIBES]-> (value)
        MERGE (descr) -[:HAS_GENE] -> (gene_context);
        """
        try:
            tx.run(query, **descriptor)
        except ServiceUnavailable as exception:
            logging.error(f"Failed to add Variant Descriptor object\nQuery: "
                          f"{query}\nDescriptor: {descriptor}")
            raise exception
        if variant_groups:
            for grp in variant_groups:
                params = descriptor.copy()
                params['group_id'] = grp['id']
                params['group_label'] = grp['label']
                params['group_description'] = grp.get('description', '')

                query = f"""
                MERGE (grp:VariationGroup {{id:$group_id, label:$group_label,
                                            description:$group_description}})
                MERGE (var:VariationDescriptor {{ {descriptor_keys} }})
                MERGE (var) -[:IN_VARIATION_GROUP]-> (grp)
                """
                try:
                    tx.run(query, **params)
                except ServiceUnavailable as exception:
                    logging.error(f"Failed to add Variant Descriptor object\n"
                                  f"Query: {query}\nDescriptor: {descriptor}")
                    raise exception

    @staticmethod
    def _add_proposition(tx, proposition: Dict):
        """Add Proposition object to DB.

        :param Dict proposition: must include `disease_context`, `therapy`,
            and `has_originating_context` fields.
        """
        formatted_keys = _create_keys_string(proposition, ('id', 'predicate',
                                                           'type'))
        prop_type = proposition.get('type')
        if prop_type == "therapeutic_response_proposition":
            prop_label = ":TherapeuticResponse"
            therapy_obj = "MERGE (therapy:Therapy {id:$object})"
            therapy_rel = """MERGE (response) -[:HAS_OBJECT]-> (therapy)
            MERGE (therapy)-[:IS_OBJECT_OF]->(response)"""
        else:
            therapy_obj, therapy_rel = "", ""
            if prop_type == "prognostic_proposition":
                prop_label = ":Prognostic"
            elif prop_type == "diagnostic_proposition":
                prop_label = ":Diagnostic"
            else:
                prop_label = ""

        query = f"""
        MERGE (response{prop_label}:Proposition
            {{ {formatted_keys} }})
        MERGE (disease:Disease {{id:$object_qualifier}})
        {therapy_obj}
        MERGE (variation:Variation {{id:$subject}})
        MERGE (response) -[:HAS_SUBJECT]-> (variation)
        MERGE (variation) -[:IS_SUBJECT_OF]-> (response)
        {therapy_rel}
        MERGE (response) -[:HAS_OBJECT_QUALIFIER]-> (disease)
        MERGE (disease) -[:IS_OBJECT_QUALIFIER_OF]-> (response)
        """

        try:
            tx.run(query, **proposition)
        except ServiceUnavailable as exception:
            logging.error(f"Failed to add Proposition object\n"
                          f"Query: {query}\nProposition: {proposition}")
            raise exception

    @staticmethod
    def _add_document(tx, document: Dict):
        """Add Document object to DB.
        :param Dict document: must include `id` field.
        """
        try:
            query = "MATCH (n:Document {id:$id}) RETURN n"
            result = tx.run(query, **document)
        except ServiceUnavailable as exception:
            logging.error(f"Failed to read Document object\n"
                          f"Query: {query}\nDocument: "
                          f"{document}")
            raise exception

        if not result.single():
            formatted_keys = _create_keys_string(document,
                                                 ('id', 'label',
                                                  'document_id',
                                                  'xrefs',
                                                  'description'))
            query = f"""
            MERGE (n:Document {{ {formatted_keys} }});
            """
            try:
                tx.run(query, **document)
            except ServiceUnavailable as exception:
                logging.error(f"Failed to add Document object\n"
                              f"Query: {query}\nDocument: "
                              f"{document}")
                raise exception

    def _get_ids_from_statement(self, statement, added_ids):
        """Add descriptors and method IDs to list of added_ids.

        :param Node statement: Statement node
        :param set added_ids: IDs found in statements
        """
        for node_id in [statement.get('therapy_descriptor'),
                        statement.get('variation_descriptor'),
                        statement.get('disease_descriptor'),
                        statement.get('method')]:
            if node_id:
                added_ids.add(node_id)

    @staticmethod
    def _add_statement(tx, statement: Dict, added_ids: Set[str]):
        """Add Statement object to DB.
        :param Dict statement: must include `id`, `variation_descriptor`,
            `disease_descriptor`, `method`, and `supported_by` fields as well
            as optional `therapy_descriptor` field
        :param set added_ids: IDs found in statements
        """
        formatted_keys = _create_keys_string(statement, ('id', 'description',
                                                         'direction',
                                                         'variation_origin',
                                                         'evidence_level'))
        match_line = ""
        rel_line = ""
        supported_by = statement.get('supported_by', [])
        if supported_by:
            for i, ev in enumerate(supported_by):
                name = f"doc_{i}"
                statement[name] = ev
                match_line += f"MERGE ({name} {{ id:${name} }})\n"
                rel_line += f"MERGE (ev) -[:CITES]-> ({name})\n"

        td = statement.get('therapy_descriptor')
        if td:
            therapy_descriptor = \
                f"MERGE (ther:TherapyDescriptor {{id:$therapy_descriptor}})"  # noqa: F541, E501
            therapy_obj = f"MERGE (ev) -[:HAS_THERAPY]-> (ther)"  # noqa: F541
            added_ids.add(td)
        else:
            therapy_descriptor, therapy_obj = "", ""

        query = f"""
        MERGE (ev:Statement {{ {formatted_keys} }})
        MERGE (prop:Proposition {{id:$proposition}})
        MERGE (var:VariationDescriptor {{id:$variation_descriptor}})
        {therapy_descriptor}
        MERGE (dis:DiseaseDescriptor {{id:$disease_descriptor}})
        MERGE (method:Method {{id:$method}})
        {match_line}
        MERGE (ev) -[:DEFINED_BY]-> (prop)
        MERGE (ev) -[:HAS_VARIATION]-> (var)
        {therapy_obj}
        MERGE (ev) -[:HAS_DISEASE]-> (dis)
        MERGE (ev) -[:USES_METHOD]-> (method)
        {rel_line}
        """

        try:
            tx.run(query, **statement)
        except ServiceUnavailable as exception:
            logging.error(f"Failed to add Evidence object\n"
                          f"Query: {query}\nEvidence: {statement}")
            raise exception

    @staticmethod
    def get_secret():
        """Get secrets for MetaKB instances."""
        secret_name = environ['METAKB_DB_PASSWORD']
        region_name = "us-east-2"

        # Create a Secrets Manager client
        session = boto3.session.Session()
        client = session.client(
            service_name='secretsmanager',
            region_name=region_name
        )

        try:
            get_secret_value_response = client.get_secret_value(
                SecretId=secret_name
            )
        except ClientError as e:
            logger.warning(e)
            if e.response['Error']['Code'] == 'DecryptionFailureException':
                # Secrets Manager can't decrypt the protected
                # secret text using the provided KMS key.
                raise e
            elif e.response['Error']['Code'] == \
                    'InternalServiceErrorException':
                # An error occurred on the server side.
                raise e
            elif e.response['Error']['Code'] == 'InvalidParameterException':
                # You provided an invalid value for a parameter.
                raise e
            elif e.response['Error']['Code'] == 'InvalidRequestException':
                # You provided a parameter value that is not valid for
                # the current state of the resource.
                raise e
            elif e.response['Error']['Code'] == 'ResourceNotFoundException':
                # We can't find the resource that you asked for.
                raise e
        else:
            # Decrypts secret using the associated KMS CMK.
            # Depending on whether the secret is a string or binary,
            # one of these fields will be populated.
            if 'SecretString' in get_secret_value_response:
                secret = get_secret_value_response['SecretString']
                return secret
            else:
                decoded_binary_secret = base64.b64decode(
                    get_secret_value_response['SecretBinary'])
                return decoded_binary_secret



"""A module for computing deltas."""
from metakb import APP_ROOT
import json
import logging
from jsondiff import diff
from datetime import date
from metakb.harvesters import CIViCHarvester, MOAHarvester
HARVESTER_CLASS = {
    'civic': CIViCHarvester,
    'moa': MOAHarvester
}
logger = logging.getLogger('metakb.delta')
logger.setLevel(logging.DEBUG)


class Delta:
    """A class for computing deltas."""

    def __init__(self, main_json, src, *args, **kwargs):
        """Initialize the Delta class.

        :param str main_json: The path to the main composite json file.
        :param str src: The source to compute the delta on
        """
        self._src = src.lower()
        assert self._src in HARVESTER_CLASS.keys()
        self._main_json = main_json
        if '_updated_json' in kwargs:
            # The path to the updated harvester composite json file.
            self._updated_json = kwargs['_updated_json']
        else:
            self._updated_json = None

    def compute_delta(self):
        """Compute delta for store computed delta in a JSON file.

        :return: A dictionary of ids to delete, update, or insert to the main
                 harvester.
        """
        # Main harvester
        with open(self._main_json, 'r') as f:
            main_json = json.load(f)

        current_date = date.today().strftime('%Y%m%d')

        # updated harvester
        if self._updated_json:
            # Updated harvester file already exists
            with open(self._updated_json, 'r') as f:
                updated_json = json.load(f)
        else:
            # Want to create updated harvester file
            fn = f"{self._src}_harvester_{current_date}.json"
            HARVESTER_CLASS[self._src]().harvest(filename=fn)
            with open(f"{APP_ROOT}/data/{self._src}/harvester/{fn}", 'r') as f:
                updated_json = json.load(f)

        delta = {
            '_meta': {
                'metakb_version': '1.0.1',
                'date_harvested': current_date
            }
        }

        if self._src == 'civic':
            delta['_meta']['civicpy_version'] = '1.1.2'
        elif self._src == 'moa':
            delta['_meta']['moa_api_version'] = '0.2'

        for record_type in main_json.keys():
            delta[record_type] = {
                'DELETE': [],
                'INSERT': [],
                'UPDATE': []
            }
            updated = updated_json[record_type]
            main = main_json[record_type]
            updated_ids = self._get_ids(updated)
            main_ids = self._get_ids(main)

            additional_ids = list(set(updated_ids) - set(main_ids))
            self._ins_del_delta(delta, record_type, 'INSERT', additional_ids,
                                updated)
            remove_ids = list(set(main_ids) - set(updated_ids))
            self._ins_del_delta(delta, record_type, 'DELETE', remove_ids,
                                main)

            self._update_delta(delta, record_type, updated, main)

        self._create_json(delta, current_date)
        return delta

    def _ins_del_delta(self, delta, record_type, key, ids_list, data):
        """Store records that will be deleted or inserted.

        :param dict delta: The deltas
        :param str record_type: The type of record
        :param str key: 'INSERT' or 'DELETE'
        :param list ids_list: A list of ids
        :param dict data: Harvester data
        """
        for record in data:
            if record['id'] in ids_list:
                delta[record_type][key].append(record)

    def _update_delta(self, delta, record_type, updated, main):
        """Store deltas.

        :param dict delta: The deltas
        :param str record_type: The type of record
        :param dict updated: updated harvester data
        :param dict main: Main harvester data
        """
        for updated_record in updated:
            for main_record in main:
                if main_record['id'] == updated_record['id']:
                    if updated_record != main_record:
                        delta[record_type]['UPDATE'].append({
                            str(main_record['id']):
                                diff(main_record, updated_record, marshal=True)
                        })
                    break

    def _get_ids(self, records):
        """Return list of ids from data.

        :param dict records: A dictionary of records
        :return: A list of ids
        """
        ids = list()
        for r in records:
            r_id = r['id']
            if r_id not in ids:
                ids.append(r_id)
        return ids

    def _create_json(self, delta, current_date):
        """Create a JSON of deltas.

        :param dict delta: A dictionary containing deltas.
        :param str current_date: The current date
        """
        src_dir = APP_ROOT / 'data' / self._src / 'delta'
        src_dir.mkdir(exist_ok=True, parents=True)

        with open(f"{src_dir}/{self._src}_deltas_{current_date}.json",
                  'w+') as f:
            json.dump(delta, f, indent=4)



"""Module for VICC normalizers."""
from typing import Optional, Tuple

from ga4gh.vrsatile.pydantic.vrs_models import VRSTypes
from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor
from variation.query import QueryHandler as VariationQueryHandler
from therapy.query import QueryHandler as TherapyQueryHandler
from therapy.schemas import NormalizationService as NormalizedTherapy
from disease.query import QueryHandler as DiseaseQueryHandler
from disease.schemas import NormalizationService as NormalizedDisease
from gene.query import QueryHandler as GeneQueryHandler
from gene.schemas import NormalizeService as NormalizedGene
import logging

logger = logging.getLogger('metakb.normalizers')
logger.setLevel(logging.DEBUG)


class VICCNormalizers:
    """A class for normalizing terms using VICC normalizers."""

    def __init__(self):
        """Initialize the VICC Normalizers."""
        self.gene_query_handler = GeneQueryHandler()
        self.variation_normalizer = VariationQueryHandler()
        self.disease_query_handler = DiseaseQueryHandler()
        self.therapy_query_handler = TherapyQueryHandler()

    def normalize_variation(self, queries) -> Optional[VariationDescriptor]:
        """Normalize variation queries.

        :param list queries: Possible query strings to try to normalize
            which are used in the event that a MANE transcript cannot be found
        :return: A normalized variation
        """
        for query in queries:
            if not query:
                continue

            try:
                variation_norm_resp = \
                    self.variation_normalizer.normalize(query)
                if variation_norm_resp:
                    if variation_norm_resp.variation.type != VRSTypes.TEXT:
                        return variation_norm_resp
            except Exception as e:  # noqa: E722
                logger.warning(f"Variation Normalizer raised an exception "
                               f"using query {query}: {e}")
        return None

    def normalize_gene(self, queries)\
            -> Tuple[Optional[NormalizedGene], Optional[str]]:
        """Normalize gene queries

        :param list queries: Gene queries to normalize
        :return: The highest matched gene's normalized response and ID
        """
        gene_norm_resp = None
        normalized_gene_id = None
        highest_match = 0
        for query_str in queries:
            if not query_str:
                continue

            gene_norm_resp = self.gene_query_handler.normalize(query_str)
            if gene_norm_resp.match_type > highest_match:
                highest_match = gene_norm_resp.match_type
                normalized_gene_id = \
                    gene_norm_resp.gene_descriptor.gene_id
                if highest_match == 100:
                    break
        return gene_norm_resp, normalized_gene_id

    def normalize_disease(self, queries)\
            -> Tuple[Optional[NormalizedDisease], Optional[str]]:
        """Normalize disease queries

        :param list queries: Disease queries to normalize
        :return: The highest matched disease's normalized response and ID
        """
        highest_match = 0
        normalized_disease_id = None
        disease_norm_resp = None

        for query in queries:
            if not query:
                continue

            disease_norm_resp = self.disease_query_handler.normalize(query)
            if disease_norm_resp.match_type > highest_match:
                highest_match = disease_norm_resp.match_type
                normalized_disease_id = \
                    disease_norm_resp.disease_descriptor.disease_id
                if highest_match == 100:
                    break
        return disease_norm_resp, normalized_disease_id

    def normalize_therapy(self, queries)\
            -> Tuple[Optional[NormalizedTherapy], Optional[str]]:
        """Normalize therapy queries

        :param list queries: Therapy queries to normalize
        :return: The highest matched therapy's normalized response and ID
        """
        highest_match = 0
        normalized_therapy_id = None
        therapy_norm_resp = None

        for query in queries:
            if not query:
                continue

            therapy_norm_resp = self.therapy_query_handler.normalize(query)
            if therapy_norm_resp.match_type > highest_match:
                highest_match = therapy_norm_resp.match_type
                normalized_therapy_id = therapy_norm_resp.therapy_descriptor.therapy_id  # noqa: E501
                if highest_match == 100:
                    break
        return therapy_norm_resp, normalized_therapy_id



"""A module for the Transform base class."""
from typing import Dict, Optional, List
import json
import logging
from pathlib import Path
from datetime import datetime as dt

from ga4gh.core import sha512t24u

from metakb import APP_ROOT, DATE_FMT
from metakb.schemas import PropositionType, Predicate, DiagnosticPredicate, \
    PrognosticPredicate, PredictivePredicate, FunctionalPredicate, \
    PathogenicPredicate
from metakb.normalizers import VICCNormalizers

logger = logging.getLogger('metakb')
logger.setLevel(logging.DEBUG)


class Transform:
    """A base class for transforming harvester data."""

    def __init__(self,
                 data_dir: Path = APP_ROOT / "data",
                 harvester_path: Optional[Path] = None,
                 normalizers: Optional[VICCNormalizers] = None) -> None:
        """Initialize Transform base class.

        :param Path data_dir: Path to source data directory
        :param Optional[Path] harvester_path: Path to previously harvested data
        :param VICCNormalizers normalizers: normalizer collection instance
        """
        self.name = self.__class__.__name__.lower().split("transform")[0]
        self.data_dir = data_dir / self.name
        self.harvester_path = harvester_path

        if normalizers is None:
            self.vicc_normalizers = VICCNormalizers()
        else:
            self.vicc_normalizers = normalizers

        self.statements = list()
        self.propositions = list()
        self.variation_descriptors = list()
        self.gene_descriptors = list()
        self.therapy_descriptors = list()
        self.disease_descriptors = list()
        self.methods = list()
        self.documents = list()

        self.next_node_id = {}

    def transform(self, *args, **kwargs):
        """Transform harvested data to the Common Data Model."""
        raise NotImplementedError

    def extract_harvester(self) -> Dict[str, List]:
        """Get harvested data from file.

        :return: Dict containing Lists of entries for each object type
        """
        if self.harvester_path is None:
            today = dt.strftime(dt.today(), DATE_FMT)
            default_fname = f"{self.name}_harvester_{today}.json"
            default_path = self.data_dir / "harvester" / default_fname
            if not default_path.exists():
                raise FileNotFoundError(
                    f"Unable to open harvest file under default filename: "
                    f"{default_path.absolute().as_uri()}"
                )
            self.harvester_path = default_path
        else:
            if not self.harvester_path.exists():
                raise FileNotFoundError(
                    f"Unable to open harvester file: {self.harvester_path}"
                )
        with open(self.harvester_path, "r") as f:
            return json.load(f)

    predicate_validation = {
        PropositionType.PREDICTIVE: PredictivePredicate,
        PropositionType.DIAGNOSTIC: DiagnosticPredicate,
        PropositionType.PROGNOSTIC: PrognosticPredicate,
        PropositionType.PATHOGENIC: PathogenicPredicate,
        PropositionType.FUNCTIONAL: FunctionalPredicate
    }

    def _get_proposition_id(
        self,
        prop_type: PropositionType,
        pred: Predicate,
        variation_ids: List[str] = [],
        disease_ids: List[str] = [],
        therapy_ids: List[str] = []
    ) -> Optional[str]:
        """Retrieve stable ID for a proposition

        :param PropositionType prop_type: type of Proposition
        :param Predicate pred: proposition predicate value
        :param str variation_id: VRS ID
        :param str disease_id: normalized disease ID
        :param str therapy_id: normalized therapy ID
        :return: proposition ID, or None if prop_type and pred conflict or
            if provided parameters cannot determine correct proposition ID
        """
        if not isinstance(pred, self.predicate_validation[prop_type]):
            msg = f"{prop_type} in query conflicts with {pred}"
            logger.error(msg)
            raise ValueError(msg)

        concept_ids = variation_ids + disease_ids + therapy_ids
        terms = [prop_type.value, pred.value] + concept_ids
        terms_sorted = sorted([t.lower() for t in terms])
        blob = json.dumps(terms_sorted).encode("ascii")
        digest = sha512t24u(blob=blob)
        return f"proposition:{digest}"

    @staticmethod
    def _get_document_id(**parameters) -> str:
        """Retrieve stable ID for a document.
        :parameters: property names and values to get ID for. Assumes values
            are strings.
        :return: identifying document ID value
        """
        params_sorted = {
            key.lower(): parameters[key].lower() for key in sorted(parameters)
        }
        blob = json.dumps(params_sorted).encode("ascii")
        return f"document:{sha512t24u(blob=blob)}"

    def create_json(self, transform_dir: Optional[Path] = None,
                    filename: Optional[str] = None) -> None:
        """Create a composite JSON for transformed data.

        :param Optional[Path] transform_dir: Path to data directory for
            transformed data
        :param Optional[str] filename: Name of transformed file
        """
        if transform_dir is None:
            transform_dir = self.data_dir / "transform"
        transform_dir.mkdir(exist_ok=True, parents=True)

        composite_dict = {
            'statements': self.statements,
            'propositions': self.propositions,
            'variation_descriptors': self.variation_descriptors,
            'gene_descriptors': self.gene_descriptors,
            'therapy_descriptors': self.therapy_descriptors,
            'disease_descriptors': self.disease_descriptors,
            'methods': self.methods,
            'documents': self.documents
        }

        today = dt.strftime(dt.today(), DATE_FMT)
        if filename is None:
            filename = f"{self.name}_cdm_{today}.json"
        out = transform_dir / filename
        with open(out, 'w+') as f:
            json.dump(composite_dict, f, indent=4)



"""Main application for FastAPI."""
from fastapi import FastAPI, Query
from fastapi.openapi.utils import get_openapi
from metakb.query import QueryHandler
from metakb.version import __version__
from metakb.schemas import SearchService, SearchIDService, \
    SearchStatementsService
from typing import Optional

app = FastAPI(
    docs_url='/api/v2',
    openapi_url='/api/v2/openapi.json',
    swagger_ui_parameters={"tryItOutEnabled": True}
)
query = QueryHandler()


def custom_openapi():
    """Generate custom fields for OpenAPI response."""
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="The VICC Meta-Knowledgebase",
        version=__version__,
        description="A search interface for cancer variant interpretations"
                    " assembled by aggregating and harmonizing across multiple"
                    " cancer variant interpretation knowledgebases.",
        routes=app.routes
    )

    openapi_schema['info']['contact'] = {
        "name": "VICC",
        "email": "help@cancervariants.org",
        "url": "https://cancervariants.org"
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi
search_summary = ("Given variation, disease, therapy, and/or gene, "
                  "return associated statements and propositions.")
search_response_description = "A response to a validly-formed query."
search_description = ("Return statements and propositions associated"
                      " to the queried concepts.")
v_description = "Variation (subject) to search"
d_description = "Disease (object qualifier) to search"
t_description = "Therapy (object) to search"
g_description = "Gene to search"
s_description = "Statement ID to search"
detail_description = "Display all descriptors, methods, and documents."


@app.get('/api/v2/search',
         summary=search_summary,
         response_description=search_response_description,
         response_model=SearchService,
         description=search_description,
         response_model_exclude_none=True)
def search(variation: Optional[str] = Query(None, description=v_description),
           disease: Optional[str] = Query(None, description=d_description),
           therapy: Optional[str] = Query(None, description=t_description),
           gene: Optional[str] = Query(None, description=g_description),
           statement_id: Optional[str] = Query(None, description=s_description),  # noqa: E501
           detail: Optional[bool] = Query(False, description=detail_description)  # noqa: E501
           ):
    """Search endpoint"""
    return query.search(variation, disease, therapy, gene, statement_id,
                        detail)


search_statements_summary = (
    "Given variation, disease, therapy, and/or gene, return associated "
    "nested statements containing propositions and descriptors.")
search_statement_response_descr = "A response to a validly-formed query."
search_statements_descr = (
    "Return nested statements associated to the queried concepts.")


@app.get('/api/v2/search/statements',
         summary=search_statements_summary,
         response_description=search_statement_response_descr,
         response_model=SearchStatementsService,
         description=search_statements_descr,
         response_model_exclude_none=True)
def get_statements(
        variation: Optional[str] = Query(None, description=v_description),
        disease: Optional[str] = Query(None, description=d_description),
        therapy: Optional[str] = Query(None, description=t_description),
        gene: Optional[str] = Query(None, description=g_description),
        statement_id: Optional[str] = Query(None, description=s_description)):
    """Return nested statements for queried concepts"""
    return query.search_statements(
        variation, disease, therapy, gene, statement_id)


id_query_desc = ("Given Meta-KB statement_id, proposition_id, descriptor_id,"
                 " document_id, or method_id return the node content.")
id_search_description = ("Return node of the queried node id.")
id_description = "Node ID to search"


@app.get('/api/v2/search/{id}',
         summary=id_query_desc,
         response_description=search_response_description,
         response_model=SearchIDService,
         description=id_search_description,
         response_model_exclude_none=True)
async def search_by_id(id: str = Query(None, description=id_description)):
    """Search by ID endpoint"""
    return query.search_by_id(id)



"""A module for the CIViC harvester."""
import logging
from typing import Optional

from civicpy import civic as civicpy

from metakb.harvesters.base import Harvester

logger = logging.getLogger('metakb.harvesters.civic')
logger.setLevel(logging.DEBUG)


class CIViCHarvester(Harvester):
    """A class for the CIViC harvester."""

    def harvest(self, filename: Optional[str] = None):
        """Retrieve and store evidence, gene, variant, and assertion
        records from CIViC in composite and individual JSON files.

        :param Optional[str] filename: File name for composite json
        :return: `True` if operation was successful, `False` otherwise.
        :rtype: bool
        """
        try:
            civicpy.load_cache(on_stale='ignore')
            evidence = self._harvest_evidence()
            genes = self._harvest_genes()
            variants = self._harvest_variants()
            assertions = self._harvest_assertions()
            self.assertions = assertions
            json_created = self.create_json(
                {
                    "evidence": evidence,
                    "genes": genes,
                    "variants": variants,
                    "assertions": assertions
                },
                filename
            )
            if not json_created:
                logger.error('CIViC Harvester was not successful.')
                return False
        except Exception as e:  # noqa: E722
            logger.error(f'CIViC Harvester was not successful: {e}')
            return False
        else:
            logger.info('CIViC Harvester was successful.')
            return True

    def _get_all_evidence(self):
        """Return all evidence item records.

        :return: All civicpy evidence item records
        """
        return civicpy.get_all_evidence()

    def _harvest_evidence(self):
        """Harvest all CIViC evidence item records.

        :return: A list of all CIViC evidence item records.
        """
        evidence_items = self._get_all_evidence()
        evidence = list()

        for ev in evidence_items:
            ev_record = \
                self._evidence_item(self._get_dict(ev), is_evidence=True)
            evidence.append(ev_record)
        return evidence

    def _get_all_genes(self):
        """Return all gene records.

        :return: All civicpy gene records
        """
        return civicpy.get_all_genes()

    def _harvest_genes(self):
        """Harvest all CIViC gene records.

        :return: A list of all CIViC gene records.
        """
        genes = self._get_all_genes()
        genes_list = list()
        for gene in genes:
            g = self._harvest_gene(self._get_dict(gene))
            genes_list.append(g)
        return genes_list

    def _get_all_variants(self):
        """Return all variant records.

        :return: All civicpy variant records
        """
        return civicpy.get_all_variants()

    def _harvest_variants(self):
        """Harvest all CIViC variant records.

        :return: A list of all CIViC variant records.
        """
        variants = self._get_all_variants()
        variants_list = list()

        for variant in variants:
            v = self._harvest_variant(self._get_dict(variant))
            variants_list.append(v)
        return variants_list

    def _get_all_assertions(self):
        """Return all assertion records.

        :return: All civicpy assertion records
        """
        return civicpy.get_all_assertions()

    def _harvest_assertions(self):
        """Harvest all CIViC assertion records.

        :return: A list of all CIViC assertion records.
        """
        assertions = self._get_all_assertions()
        assertions_list = list()

        for assertion in assertions:
            a = self._harvest_assertion(self._get_dict(assertion))
            assertions_list.append(a)
        return assertions_list

    def _harvest_gene(self, gene):
        """Harvest an individual CIViC gene record.

        :param Gene gene: A CIViC gene object
        :return: A dictionary containing CIViC gene data
        """
        g = {
            'id': gene['id'],
            'name': gene['name'],
            'entrez_id': gene['entrez_id'],
            'description': gene['description'],
            'variants': [
                {
                    'name': self._get_dict(variant)['name'],
                    'id': self._get_dict(variant)['id'],
                    'evidence_items':
                        self._get_dict(variant)['_evidence_items']
                }
                for variant in gene['_variants']
            ],
            'aliases': gene['aliases'],
            # TODO: Add lifecycle_actions, sources
            'type': gene['type']
        }

        for v in g['variants']:
            evidence_items = {
                'accepted_count': 0,
                'rejected_count': 0,
                'submitted_count': 0
            }
            for e in v['evidence_items']:
                e = self._get_dict(e)
                if e['status'] == 'submitted':
                    evidence_items['submitted_count'] += 1
                elif e['status'] == 'rejected':
                    evidence_items['rejected_count'] += 1
                elif e['status'] == 'accepted':
                    evidence_items['accepted_count'] += 1
            v['evidence_items'] = evidence_items

        return g

    def _harvest_variant(self, variant):
        """Harvest an individual CIViC variant record.

        :param Gene variant: A CIViC variant object
        :return: A dictionary containing CIViC variant data
        """
        v = self._variant(variant)

        # Add more attributes to variant data
        v_extra = {
            'evidence_items': [
                self._evidence_item(self._get_dict(evidence_item))
                for evidence_item in variant['_evidence_items']
            ],
            'variant_groups': [
                {
                    'id': self._get_dict(variant_group)['id'],
                    'name': self._get_dict(variant_group)['name'],
                    'description':
                        self._get_dict(variant_group)['description'],
                    'variants': [
                        self._variant(self._get_dict(variant))
                        for variant in self._get_dict(variant_group)[
                            'variants']
                    ],
                    'type': self._get_dict(variant_group)['type']
                }
                for variant_group in variant['variant_groups']
            ],
            'assertions': [
                self._assertion(self._get_dict(assertion))
                for assertion in variant['_assertions']
            ],
            'variant_aliases': variant['variant_aliases'],
            'hgvs_expressions': variant['hgvs_expressions'],
            'clinvar_entries': variant['clinvar_entries'],
            # TODO: Add lifecycle_actions
            'allele_registry_id': variant['allele_registry_id'],
            # TODO: Add allele_registry_hgvs
        }
        v.update(v_extra)
        return v

    def _harvest_assertion(self, assertion):
        """Harvest an individual CIViC assertion record.

        :param Gene assertion: A CIViC variant object
        :return: A dictionary containing CIViC assertion data
        """
        a = self._assertion(assertion)

        # Add more attributes to assertion data
        a_extra = {
            'nccn_guideline': assertion['nccn_guideline'],
            'nccn_guideline_version': assertion['nccn_guideline_version'],
            'amp_level': assertion['amp_level'],
            'evidence_items': [
                self._evidence_item(self._get_dict(evidence_item),
                                    is_assertion=True)
                for evidence_item in assertion['evidence_items']
            ],
            'acmg_codes': assertion['acmg_codes'],
            'drug_interaction_type': assertion['drug_interaction_type'],
            'fda_companion_test': assertion['fda_companion_test'],
            'allele_registry_id': assertion['allele_registry_id'],
            'phenotypes': assertion['phenotypes'],
            'variant_origin': assertion['variant_origin']
            # TODO: Add lifecycle_actions
        }
        a.update(a_extra)
        return a

    def _evidence_item(self, evidence_item,
                       is_evidence=False, is_assertion=False):
        """Get evidence item data.

        :param Evidence evidence_item: A CIViC Evidence record
        :param bool is_evidence: Whether or not the evidence item is
                                 being harvested in an evidence record
        :param bool is_assertion: Whether or not the evidence item is
                                  being harvested in an assertion record
        :return: A dictionary containing evidence item data
        """
        e = {
            'id': evidence_item['id'],
            'name': evidence_item['name'],
            'description': evidence_item['description'],
            'disease': self._disease(self._get_dict(evidence_item)),
            'drugs': [
                self._drug(self._get_dict(drug))
                for drug in evidence_item['drugs']
            ],
            'rating': evidence_item['rating'],
            'evidence_level': evidence_item['evidence_level'],
            'evidence_type': evidence_item['evidence_type'],
            'clinical_significance':
                evidence_item['clinical_significance'],
            'evidence_direction':
                evidence_item['evidence_direction'],
            'variant_origin': evidence_item['variant_origin'],
            'drug_interaction_type':
                evidence_item['drug_interaction_type'],
            'status': evidence_item['status'],
            # TODO: Add open_change_count
            'type': evidence_item['type'],
            'source': self._source(evidence_item),
            'variant_id': evidence_item['variant_id'],
            # TODO: Find variant w phenotypes
            'phenotypes': []
        }

        # Assertions and Evidence Items contain more attributes
        if is_assertion or is_evidence:
            e['assertions'] = [
                self._assertion(self._get_dict(assertion))
                for assertion in evidence_item['_assertions']
            ]
            # TODO: Add lifecycle_actions, fields_with_pending_changes
            e['gene_id'] = evidence_item['gene_id']
            if is_assertion:
                # TODO: Add state_params
                pass

        return e

    def _variant(self, variant):
        """Get basic variant data.

        :param Variant variant: A CIViC Variant record
        :return: A dictionary containing variant data
        """
        return {
            'id': variant['id'],
            'entrez_name': variant['entrez_name'],
            'entrez_id': variant['entrez_id'],
            'name': variant['name'],
            'description': variant['description'],
            'gene_id': variant['gene_id'],
            'type': variant['type'],
            'variant_types': [
                self._variant_types(self._get_dict(variant_type))
                for variant_type in variant['variant_types']
            ],
            'civic_actionability_score':
                int(variant['civic_actionability_score']) if int(variant['civic_actionability_score']) == variant['civic_actionability_score'] else variant['civic_actionability_score'],  # noqa: E501
            'coordinates':
                self._variant_coordinates(variant)
        }

    def _assertion(self, assertion):
        """Get assertion data.

        :param Assertion assertion: A CIViC Assertion record
        :return: A dictionary containing assertion data
        """
        disease = self._get_dict(assertion['disease'])
        return {
            'id': assertion['id'],
            'type': assertion['type'],
            'name': assertion['name'],
            'summary': assertion['summary'],
            'description': assertion['description'],
            'gene': self._gene_name_id(assertion),
            'variant': self._variant_name_id(assertion),
            'disease': {
                'id': disease['id'],
                'name': disease['name'],
                'display_name': disease['display_name'],
                'doid': disease['doid'],
                'url': disease['url']
            },
            'drugs': [
                self._drug(self._get_dict(drug))
                for drug in assertion['drugs']
            ],
            'evidence_type': assertion['evidence_type'],
            'evidence_direction': assertion['evidence_direction'],
            'clinical_significance': assertion['clinical_significance'],
            # TODO: Add evidence_item_count
            'fda_regulatory_approval':
                assertion['fda_regulatory_approval'],
            'status': assertion['status'],
            # TODO: Add open_change_count, pending_evidence_count
        }

    def _variant_coordinates(self, variant):
        """Get a variant's coordinates.

        :param Variant variant: A CIViC variant record
        :return: A dictionary containing a variant's coordinates
        """
        coordinates = self._get_dict(variant['coordinates'])
        return {
            'chromosome': coordinates['chromosome'],
            'start': coordinates['start'],
            'stop': coordinates['stop'],
            'reference_bases': coordinates['reference_bases'],
            'variant_bases': coordinates['variant_bases'],
            'representative_transcript':
                coordinates['representative_transcript'],
            'chromosome2': coordinates['chromosome2'],
            'start2': coordinates['start2'],
            'stop2': coordinates['stop2'],
            'representative_transcript2':
                coordinates['representative_transcript2'],
            'ensembl_version': coordinates['ensembl_version'],
            'reference_build': coordinates['reference_build']
        }

    def _variant_types(self, variant_type):
        """Get variant_type data.

        :param CivicAttribute variant_type: A CIViC variant_type record
        :return: A dictionary containing variant_type data
        """
        return {
            'id': variant_type['id'],
            'name': variant_type['name'],
            'display_name': variant_type['display_name'],
            'so_id': variant_type['so_id'],
            'description': variant_type['description'],
            'url': variant_type['url']
        }

    def _source(self, evidence_item):
        """Get an evidence item's source data.

        :param Evidence evidence_item: A CIViC Evidence record
        :return: A dictionary containing source data
        """
        source = self._get_dict(evidence_item['source'])
        return {
            'id': source['id'],
            'name': source['name'],
            'citation': source['citation'],
            'citation_id': source['citation_id'],
            'source_type': source['source_type'],
            'asco_abstract_id': source['asco_abstract_id'],
            'source_url': source['source_url'],
            'open_access': source['open_access'],
            'pmc_id': source['pmc_id'],
            'publication_date': source['publication_date'],
            'journal': source['journal'],
            'full_journal_title': source['full_journal_title'],
            'status': source['status'],
            'is_review': source['is_review'],
            'clinical_trials': [ct for ct in source['clinical_trials']]
        }

    def _disease(self, evidence_item):
        """Get an evidence item's disease data.
        :param Evidence evidence_item: A CIViC Evidence record
        :return: A dictionary containing disease data
        """
        disease = self._get_dict(evidence_item['disease'])
        if not disease:
            return None
        else:
            return {
                'id': disease['id'],
                'name': disease['name'],
                'display_name': disease['display_name'],
                'doid': disease['doid'],
                'url': disease['url']
            }

    def _drug(self, drug):
        """Get drug data.

        :param Drug drug: A CIViC Drug record
        :return: A dictionary containing drug data.
        """
        drug = self._get_dict(drug)
        return {
            "id": drug['id'],
            "name": drug['name'],
            "ncit_id": drug['ncit_id'],
            "aliases": drug['aliases']
        }

    def _gene_name_id(self, assertion):
        """Get gene name and id.

        :param Assertion assertion: A CIViC Assertion record
        :return: A dictionary containing a gene's name and id
        """
        gene = self._get_dict(assertion['gene'])
        return {
            'name': gene['name'],
            'id': gene['id']
        }

    def _variant_name_id(self, assertion):
        """Get variant name and id.

        :param Assertion assertion: A CIViC Assertion record
        :return: A dictionary containing a variant's name and id
        """
        variant = self._get_dict(assertion['variant'])
        return {
            'name': variant['name'],
            'id': variant['id']
        }

    def _get_dict(self, obj):
        """Return the __dict__ attribute for an object.

        :param obj: The civicpy object
        :return: A dictionary for the object
        """
        if isinstance(obj, (civicpy.Drug, civicpy.Disease,
                            civicpy.CivicAttribute, civicpy.Evidence,
                            civicpy.CivicRecord, civicpy.Gene,
                            civicpy.Assertion, civicpy.Variant)):
            return vars(obj)
        else:
            return obj
